@@ -7,9 +7,11 @@
 PITCH_RANGE = range(21, 109)  # the recommended pitches for piano in the GM2 specs are from 21 to 108
 BEAT_RES = {(0, 4): 8, (4, 12): 4}  # samples per beat
 NB_VELOCITIES = 32  # nb of velocity bins, velocities values from 0 to 127 will be quantized
-ADDITIONAL_TOKENS = {'Chord': True,
-                     'Empty': True,
-                     'Tempo': False,  # Unused for now (not implemented)
+ADDITIONAL_TOKENS = {'Chord': False,
+                     'Rest': False,
+                     'Tempo': False,
+                     # rest params
+                     'rest_range': (2, 8),  # (/min_rest, max_rest_in_BEAT), first divides a whole note/rest
                      # tempo params
                      'nb_tempos': 32,  # nb of tempo bins for additional tempo tokens, quantized like velocities
                      'tempo_range': (40, 250)}  # (min_tempo, max_tempo)
