@@ -8,7 +8,7 @@ from typing import List, Tuple, Dict, Optional
 import numpy as np
 from miditoolkit import Instrument, Note, TempoChange
 
-from .midi_tokenizer_base import MIDITokenizer, Event, detect_chords
+from .midi_tokenizer_base import MIDITokenizer, Vocabulary, Event, detect_chords
 from .constants import *
 
 
@@ -30,18 +30,19 @@ class MIDILikeEncoding(MIDITokenizer):
     :param beat_res: beat resolutions, with the form:
             {(beat_x1, beat_x2): beat_res_1, (beat_x2, beat_x3): beat_res_2, ...}
             The keys of the dict are tuples indicating a range of beats, ex 0 to 3 for the first bar
-            The values are the resolution, in frames per beat, of the given range, ex 8
+            The values are the resolution, in samples per beat, of the given range, ex 8
     :param nb_velocities: number of velocity bins
-    :param additional_tokens: specifies additional tokens (chords, empty bars, tempo...)
+    :param additional_tokens: specifies additional tokens (chords, time signature, rests, tempo...)
     :param program_tokens: will add entries for MIDI programs in the dictionary, to use
             in the case of multitrack generation for instance
+    :param sos_eos_tokens: Adds Start Of Sequence (SOS) and End Of Sequence (EOS) tokens to the vocabulary
     :param params: can be a path to the parameter (json encoded) file or a dictionary
     """
     def __init__(self, pitch_range: range = PITCH_RANGE, beat_res: Dict[Tuple[int, int], int] = BEAT_RES,
                  nb_velocities: int = NB_VELOCITIES, additional_tokens: Dict[str, bool] = ADDITIONAL_TOKENS,
-                 program_tokens: bool = PROGRAM_TOKENS, params=None):
-        additional_tokens['Empty'] = False  # Incompatible additional tokens
-        super().__init__(pitch_range, beat_res, nb_velocities, additional_tokens, program_tokens, params)
+                 program_tokens: bool = PROGRAM_TOKENS, sos_eos_tokens: bool = False, params=None):
+        super().__init__(pitch_range, beat_res, nb_velocities, additional_tokens,
+                         {'program_tokens': program_tokens, 'sos_eos_tokens': sos_eos_tokens}, params)
 
     def track_to_tokens(self, track: Instrument) -> List[int]:
         """ Converts a track (miditoolkit.Instrument object) into a sequence of tokens
@@ -52,65 +53,68 @@ class MIDILikeEncoding(MIDITokenizer):
         """
         # Make sure the notes are sorted first by their onset (start) times, second by pitch
         # notes.sort(key=lambda x: (x.start, x.pitch))  # done in midi_to_tokens
+        ticks_per_sample = self.current_midi_metadata['time_division'] / max(self.beat_res.values())
         events = []
 
         # Creates the Note On, Note Off and Velocity events
         for n, note in enumerate(track.notes):
-            if note.pitch not in self.pitch_range:  # Notes to low or to high are discarded
-                continue
             # Note On
-            events.append(Event(
-                name='Note-On',
-                time=note.start,
-                value=note.pitch,
-                text=note.pitch))
+            events.append(Event(type_='Note-On', time=note.start, value=note.pitch, desc=note.end))
             # Velocity
-            velocity_index = (np.abs(self.velocity_bins - note.velocity)).argmin()
-            events.append(Event(
-                name='Velocity',
-                time=note.start,
-                value=velocity_index,
-                text=f'{note.velocity}/{self.velocity_bins[velocity_index]}'))
+            events.append(Event(type_='Velocity', time=note.start, value=note.velocity, desc=f'{note.velocity}'))
             # Note Off
-            events.append(Event(
-                name='Note-Off',
-                time=note.end,
-                value=note.pitch,
-                text=note.pitch))
+            events.append(Event(type_='Note-Off', time=note.end, value=note.pitch, desc=note.end))
         # Adds tempo events if specified
         if self.additional_tokens['Tempo']:
             for tempo_change in self.current_midi_metadata['tempo_changes']:
-                events.append(Event(
-                    name='Tempo',
-                    time=tempo_change.time,
-                    value=(np.abs(self.tempo_bins - tempo_change.tempo)).argmin(),
-                    text=tempo_change.tempo))
+                events.append(Event(type_='Tempo', time=tempo_change.time, value=tempo_change.tempo,
+                                    desc=tempo_change.tempo))
 
         # Sorts events in the good order
         events.sort(key=lambda x: x.time)
 
         # Time Shift
-        current_tick = 0
-        for e, event in enumerate(events[:-1].copy()):
-            if event.time == current_tick:
+        previous_tick = 0
+        previous_note_end = track.notes[0].start + 1
+        for e, event in enumerate(events.copy()):
+            if event.type == 'Note-On':
+                previous_note_end = max(previous_note_end, event.desc)
+
+            if event.time == previous_tick:
                 continue
-            time_shift = event.time - current_tick
-            index = np.argmin(np.abs([ticks - time_shift for ticks in
-                                      self.durations_ticks[self.current_midi_metadata['time_division']]]))
-            events.append(Event(
-                name='Time-Shift',
-                time=current_tick,
-                value='.'.join(map(str, self.durations[index])),
-                text=f'{time_shift} ticks'))
-            current_tick = event.time
+
+            time_shift = event.time - previous_tick
+
+            if self.additional_tokens['Rest'] and event.time > previous_note_end:
+                rest_beat, rest_pos = divmod(time_shift, self.current_midi_metadata['time_division'])
+                rest_beat = min(rest_beat, max([r[0] for r in self.rests]))
+                rest_pos = round(rest_pos / ticks_per_sample)
+                if rest_beat > 0:
+                    events.append(Event(type_='Rest', time=previous_tick, value=f'{rest_beat}.0',
+                                        desc=f'{rest_beat}.0'))
+                    previous_tick += rest_beat * self.current_midi_metadata['time_division']
+
+                while rest_pos >= self.rests[0][1]:
+                    rest_pos_temp = min([r[1] for r in self.rests], key=lambda x: abs(x - rest_pos))
+                    events.append(Event(type_='Rest', time=previous_tick, value=f'0.{rest_pos_temp}',
+                                        desc=f'0.{rest_pos_temp}'))
+                    previous_tick += round(rest_pos_temp * ticks_per_sample)
+                    rest_pos -= rest_pos_temp
+
+            else:
+                index = np.argmin(np.abs([ticks - time_shift for ticks in
+                                          self.durations_ticks[self.current_midi_metadata['time_division']]]))
+                events.append(Event(type_='Time-Shift', time=previous_tick,
+                                    value='.'.join(map(str, self.durations[index])), desc=f'{time_shift} ticks'))
+            previous_tick = event.time
 
         # Adds chord events if specified
         if self.additional_tokens['Chord'] and not track.is_drum:
-            events += detect_chords(track.notes, self.current_midi_metadata['time_division'])
+            events += detect_chords(track.notes, self.current_midi_metadata['time_division'], self._first_beat_res)
 
         events.sort(key=lambda x: (x.time, self._order(x)))
 
-        return self.events_to_tokens(events)
+        return self._events_to_tokens(events)
 
     def tokens_to_track(self, tokens: List[int], time_division: Optional[int] = TIME_DIVISION,
                         program: Optional[Tuple[int, bool]] = (0, False), default_duration: int = None) \
@@ -124,35 +128,36 @@ class MIDILikeEncoding(MIDITokenizer):
                                 note off event. Leave None to discard Note On with no Note Off event.
         :return: the miditoolkit instrument object and tempo changes
         """
-        events = self.tokens_to_events(tokens)
+        ticks_per_sample = time_division // max(self.beat_res.values())
+        events = self._tokens_to_events(tokens)
 
         max_duration = (self.durations[-1][0] + self.durations[-1][1]) * time_division
         name = 'Drums' if program[1] else MIDI_INSTRUMENTS[program[0]]['name']
         instrument = Instrument(program[0], is_drum=program[1], name=name)
-        if self.additional_tokens['Tempo']:
-            tempo_changes = [TempoChange(TEMPO, -1)]  # mock the first tempo change to optimize below
-        else:  # default
-            tempo_changes = [TempoChange(TEMPO, 0)] * 2  # the first will be deleted at the end of the method
+        tempo_changes = [TempoChange(TEMPO, -1)]  # mock the first tempo change to optimize below
 
         current_tick = 0
-        count = 0
-        while count < len(events):
-            if events[count].name == 'Note-On':
+        ei = 0
+        while ei < len(events):
+            if events[ei].type == 'Note-On':
                 try:
-                    if events[count + 1].name == 'Velocity':
-                        pitch = int(events[count].value)
-                        vel = int(self.velocity_bins[int(events[count + 1].value)])
+                    if events[ei + 1].type == 'Velocity':
+                        pitch = int(events[ei].value)
+                        vel = int(events[ei + 1].value)
 
                         # look for an associated note off event to get duration
                         offset_tick = 0
                         duration = 0
-                        for i in range(count+1, len(events)):
-                            if events[i].name == 'Note-Off' and int(events[i].value) == pitch:
+                        for i in range(ei+1, len(events)):
+                            if events[i].type == 'Note-Off' and int(events[i].value) == pitch:
                                 duration = offset_tick
                                 break
-                            elif events[i].name == 'Time-Shift':
+                            elif events[i].type == 'Time-Shift':
                                 beat, pos, res = map(int, events[i].value.split('.'))
                                 offset_tick += (beat * res + pos) * time_division // res
+                            elif events[ei].type == 'Rest':
+                                beat, pos = map(int, events[ei].value.split('.'))
+                                current_tick += beat * time_division + pos * ticks_per_sample
                             if offset_tick > max_duration:  # will not look for Note Off beyond
                                 break
 
@@ -160,95 +165,79 @@ class MIDILikeEncoding(MIDITokenizer):
                             duration = default_duration
                         if duration != 0:
                             instrument.notes.append(Note(vel, pitch, current_tick, current_tick + duration))
-                        count += 2
+                        ei += 1
                 except IndexError as _:
-                    count += 1
-            elif events[count].name == 'Time-Shift':
-                beat, pos, res = map(int, events[count].value.split('.'))
+                    pass
+            elif events[ei].type == 'Time-Shift':
+                beat, pos, res = map(int, events[ei].value.split('.'))
                 current_tick += (beat * res + pos) * time_division // res
-                count += 1
-            elif events[count].name == 'Tempo':
-                tempo = int(self.tempo_bins[int(events[count].value)])
+            elif events[ei].type == 'Rest':
+                beat, pos = map(int, events[ei].value.split('.'))
+                current_tick += beat * time_division + pos * ticks_per_sample
+            elif events[ei].type == 'Tempo':
+                tempo = int(events[ei].value)
                 if tempo != tempo_changes[-1].tempo:
                     tempo_changes.append(TempoChange(tempo, current_tick))
-                count += 1
-            else:
-                count += 1
-        del tempo_changes[0]
+            ei += 1
+        if len(tempo_changes) > 1:
+            del tempo_changes[0]  # delete mocked tempo change
+        tempo_changes[0].time = 0
         return instrument, tempo_changes
 
-    def _create_vocabulary(self, program_tokens: bool) -> Tuple[dict, dict, dict]:
-        """ Create the tokens <-> event dictionaries
-        These dictionaries are created arbitrary according to constants defined
-        at the top of this file.
-        Note that when using them (prepare_data method), there is no error-handling
-        so you must be sure that every case is covered by the dictionaries.
-        NOTE: token index 0 is often used as a padding index during training, it might
-        be preferable to leave it as it
+    def _create_vocabulary(self, program_tokens: bool, sos_eos_tokens: bool = False) -> Vocabulary:
+        """ Creates the Vocabulary object of the tokenizer.
+        See the docstring of the Vocabulary class for more details about how to use it.
+        NOTE: token index 0 is often used as a padding index during training
 
-        :param program_tokens: creates tokens for MIDI programs in the dictionary
-        :return: the dictionaries, one for each translation
+        :param program_tokens: will include tokens for MIDI programs
+        :param sos_eos_tokens: will include Start Of Sequence (SOS) and End Of Sequence (tokens)
+        :return: the vocabulary object
         """
-        event_to_token = {'PAD_None': 0}  # starting at 1, token 0 is for padding
-        token_type_indices = {'Pad': [0]}  # Empty is for empty bars
-        count = 1
+        vocab = Vocabulary({'PAD_None': 0})
 
         # NOTE ON
-        token_type_indices['Note-On'] = list(range(count, count + len(self.pitch_range)))
-        for i in self.pitch_range:
-            event_to_token[f'Note-On_{i}'] = count
-            count += 1
+        vocab.add_event(f'Note-On_{i}' for i in self.pitch_range)
 
         # NOTE OFF
-        token_type_indices['Note-Off'] = list(range(count, count + len(self.pitch_range)))
-        for i in self.pitch_range:
-            event_to_token[f'Note-Off_{i}'] = count
-            count += 1
+        vocab.add_event(f'Note-Off_{i}' for i in self.pitch_range)
 
         # VELOCITY
-        token_type_indices['Velocity'] = list(range(count, count + len(self.velocity_bins)))
-        for i in range(len(self.velocity_bins)):
-            event_to_token[f'Velocity_{i}'] = count
-            count += 1
+        vocab.add_event(f'Velocity_{i}' for i in self.velocities)
 
         # TIME SHIFTS
-        token_type_indices['Time-Shift'] = list(range(count, count + len(self.durations)))
-        for i in range(0, len(self.durations)):
-            event_to_token[f'Time-Shift_{".".join(map(str, self.durations[i]))}'] = count
-            count += 1
+        vocab.add_event(f'Time-Shift_{".".join(map(str, self.durations[i]))}' for i in range(len(self.durations)))
 
         # CHORD
         if self.additional_tokens['Chord']:
-            token_type_indices['Chord'] = list(range(count, count + 3 + len(CHORD_MAPS)))
-            for i in range(3, 6):  # non recognized chords, just considers the nb of notes (between 3 and 5 only)
-                event_to_token[f'Chord_{i}'] = count
-                count += 1
-            for chord_quality in CHORD_MAPS:  # classed chords
-                event_to_token[f'Chord_{chord_quality}'] = count
-                count += 1
+            vocab.add_event(f'Chord_{i}' for i in range(3, 6))  # non recognized chords (between 3 and 5 notes only)
+            vocab.add_event(f'Chord_{chord_quality}' for chord_quality in CHORD_MAPS)
+
+        # REST
+        if self.additional_tokens['Rest']:
+            vocab.add_event(f'Rest_{".".join(map(str, rest))}' for rest in self.rests)
 
         # TEMPO
         if self.additional_tokens['Tempo']:
-            token_type_indices['Tempo'] = list(range(count, count + len(self.tempo_bins)))
-            for i in range(len(self.tempo_bins)):
-                event_to_token[f'Tempo_{i}'] = count
-                count += 1
+            vocab.add_event(f'Tempo_{i}' for i in self.tempos)
 
         # PROGRAM
         if program_tokens:
-            token_type_indices['Program'] = list(range(count, count + 129))
-            for program in range(-1, 128):  # -1 is drums
-                event_to_token[f'Program_{program}'] = count
-                count += 1
+            vocab.add_event(f'Program_{program}' for program in range(-1, 128))
 
-        token_to_event = {v: k for k, v in event_to_token.items()}  # inversion
-        return event_to_token, token_to_event, token_type_indices
+        # SOS & EOS
+        if sos_eos_tokens:
+            vocab.add_sos_eos_to_vocab()
 
-    def _create_token_types_graph(self) -> Dict[str, List[str]]:
+        return vocab
+
+    def create_token_types_graph(self) -> Dict[str, List[str]]:
         dic = dict()
 
-        if 'Program' in self.token_types_indices:
+        try:
+            _ = self.vocab.tokens_of_type('Program')
             dic['Program'] = ['Note-On', 'Time-Shift']
+        except KeyError:
+            pass
 
         dic['Note-On'] = ['Velocity']
         dic['Velocity'] = ['Note-On', 'Time-Shift']
@@ -262,10 +251,16 @@ class MIDILikeEncoding(MIDITokenizer):
 
         if self.additional_tokens['Tempo']:
             dic['Time-Shift'] += ['Tempo']
-            dic['Tempo'] = ['Time-Shift', 'Note-On']
+            dic['Tempo'] = ['Note-On']
             if self.additional_tokens['Chord']:
                 dic['Tempo'] += ['Chord']
 
+        if self.additional_tokens['Rest']:
+            dic['Rest'] = ['Rest', 'Note-On']
+            if self.additional_tokens['Chord']:
+                dic['Rest'] += ['Chord']
+            dic['Note-Off'] = ['Rest']
+
         return dic
 
     @staticmethod
@@ -275,15 +270,15 @@ class MIDILikeEncoding(MIDITokenizer):
         :param x: event to get order index
         :return: an order int
         """
-        if x.name == "Program":
+        if x.type == 'Program':
             return 0
-        elif x.name == "Note-Off":
+        elif x.type == 'Note-Off':
             return 1
-        elif x.name == 'Tempo':
+        elif x.type == 'Tempo':
             return 2
-        elif x.name == "Chord":
+        elif x.type == "Chord":
             return 3
-        elif x.name == "Time-Shift":
+        elif x.type == 'Time-Shift' or x.type == 'Rest':
             return 1000  # always last
         else:  # for other types of events, the order should be handle when inserting the events in the sequence
             return 4
