@@ -1,37 +1,22 @@
 """ MIDI encoding base class and methods
 TODO Control change messages (sustain, modulation, pitch bend)
+TODO time signature changes tokens
 
 """
 
 from sys import stdout
 from pathlib import Path, PurePath
 import json
+from collections import Counter
 from typing import List, Tuple, Dict, Union, Callable, Optional, Any
 
 import numpy as np
-from miditoolkit import MidiFile, Instrument, Note, TempoChange
+from miditoolkit import MidiFile, Instrument, Note, TempoChange, TimeSignature
 
+from .vocabulary import Vocabulary, Event
 from .constants import TIME_DIVISION, CHORD_MAPS
 
 
-class Event:
-    """ Event class, representing a token and its characteristics
-    The name corresponds to the token type (e.g. Pitch, Position ...);
-    The value to its value.
-    These two attributes are used in events_to_tokens and tokens_to_events
-    methods (see below) to convert an Event object to its corresponding integer.
-    """
-
-    def __init__(self, name, time, value, text):
-        self.name = name
-        self.time = time
-        self.value = value
-        self.text = text
-
-    def __repr__(self):
-        return f'Event(name={self.name}, time={self.time}, value={self.value}, text={self.text})'
-
-
 class MIDITokenizer:
     """ MIDI encoding base class, containing common parameters to all encodings
     and common methods.
@@ -40,16 +25,17 @@ class MIDITokenizer:
     :param beat_res: beat resolutions, with the form:
             {(beat_x1, beat_x2): beat_res_1, (beat_x2, beat_x3): beat_res_2, ...}
             The keys of the dict are tuples indicating a range of beats, ex 0 to 3 for the first bar
-            The values are the resolution, in frames per beat, of the given range, ex 8
+            The values are the resolution, in samples per beat, of the given range, ex 8
     :param nb_velocities: number of velocity bins
-    :param additional_tokens: specifies additional tokens (chords, empty bars, tempo...)
-    :param program_tokens: will add entries for MIDI programs in the dictionary, to use
-            in the case of multitrack generation for instance
+    :param additional_tokens: specifies additional tokens (chords, rests, tempo...)
+    :param vocab_args: arguments for the _create_vocabulary method
     :param params: can be a path to the parameter (json encoded) file or a dictionary
     """
+
     def __init__(self, pitch_range: range, beat_res: Dict[Tuple[int, int], int], nb_velocities: int,
-                 additional_tokens: Dict[str, bool], program_tokens: bool,
+                 additional_tokens: Dict[str, Union[bool, int, Tuple[int, int]]], vocab_args: Dict[str, any],
                  params: Union[str, Path, PurePath, Dict[str, Any]] = None):
+        # Initialize params
         if params is None:
             self.pitch_range = pitch_range
             self.beat_res = beat_res
@@ -58,16 +44,31 @@ class MIDITokenizer:
         else:
             self.load_params(params)
 
-        self.durations = self._create_durations_tuples()
-        self.velocity_bins = np.linspace(0, 127, self.nb_velocities + 1, dtype=np.intc)
-        np.delete(self.velocity_bins, 0)  # removes velocity 0
-        if additional_tokens['Tempo']:
-            self.tempo_bins = np.linspace(*additional_tokens['tempo_range'], additional_tokens['nb_tempos'],
-                                          dtype=np.intc)
-        else:
-            self.tempo_bins = np.zeros(1)
+        # Init duration and velocity values
+        self.durations = self.__create_durations_tuples()
+        self.velocities = list(np.linspace(0, 127, self.nb_velocities + 1, dtype=np.intc))
+        del self.velocities[0]  # removes velocity 0
+        self._first_beat_res = list(beat_res.values())[0]
+        for beat_range, res in beat_res.items():
+            if 0 in beat_range:
+                self._first_beat_res = res
+                break
 
-        self.event2token, self.token2event, self.token_types_indices = self._create_vocabulary(program_tokens)
+        # Tempos
+        self.tempos = np.zeros(1)
+        if additional_tokens['Tempo']:
+            self.tempos = np.linspace(*additional_tokens['tempo_range'], additional_tokens['nb_tempos'],
+                                      dtype=np.intc)
+
+        # Rests
+        self.rests = []
+        if additional_tokens['Rest']:
+            assert additional_tokens['rest_range'][0] // 4 <= self._first_beat_res, \
+                'The minimum rest value must be equal or superior to the initial beat resolution'
+            self.rests = self.__create_rests()
+
+        # Vocabulary
+        self.vocab = self._create_vocabulary(**vocab_args)
 
         # Keep in memory durations in ticks for seen time divisions so these values
         # are not calculated each time a MIDI is processed
@@ -79,8 +80,7 @@ class MIDITokenizer:
 
     def midi_to_tokens(self, midi: MidiFile) -> List[List[Union[int, List[int]]]]:
         """ Converts a MIDI file in a tokens representation.
-        NOTE: if you override this method, be sure to keep every line of code below until
-        the "Convert track to token" comment in the for loop
+        NOTE: if you override this method, be sure to keep the first lines in your method
 
         :param midi: the MIDI objet to convert
         :return: the token representation, i.e. tracks converted into sequences of tokens
@@ -92,26 +92,48 @@ class MIDITokenizer:
             self.durations_ticks[midi.ticks_per_beat] = [(beat * res + pos) * midi.ticks_per_beat // res
                                                          for beat, pos, res in self.durations]
 
+        # Preprocess the MIDI file
+        self.preprocess_midi(midi)
+
         # Register MIDI metadata
         self.current_midi_metadata = {'time_division': midi.ticks_per_beat,
                                       'tempo_changes': midi.tempo_changes,
                                       'time_sig_changes': midi.time_signature_changes,
                                       'key_sig_changes': midi.key_signature_changes}
 
-        # Quantize tempo changes times
-        quantize_tempos(midi.tempo_changes, midi.ticks_per_beat, max(self.beat_res.values()))
+        # **************** OVERRIDE FROM HERE, KEEP THE LINES ABOVE IN YOUR METHOD ****************
 
-        tokens = []
-        for track in midi.instruments:
-            quantize_note_times(track.notes, self.current_midi_metadata['time_division'], max(self.beat_res.values()))
-            track.notes.sort(key=lambda x: (x.start, x.pitch))  # sort notes
-            remove_duplicated_notes(track.notes)  # remove possible duplicated notes
-
-            # Convert track to tokens
-            tokens.append(self.track_to_tokens(track))
+        # Convert each track to tokens
+        tokens = [self.track_to_tokens(track) for track in midi.instruments]
 
         return tokens
 
+    def preprocess_midi(self, midi: MidiFile):
+        """ Will process a MIDI file so it can be used to train a model.
+        Its notes attributes (times, pitches, velocities) will be quantized and sorted, duplicated
+        notes removed, as well as tempos.
+        NOTE: empty tracks (with no note) will be removed from the MIDI object
+
+        :param midi: MIDI object to preprocess
+        """
+        t = 0
+        while t < len(midi.instruments):
+            self.quantize_notes(midi.instruments[t].notes, midi.ticks_per_beat)  # quantize notes attributes
+            midi.instruments[t].notes.sort(key=lambda x: (x.start, x.pitch, x.end))  # sort notes
+            remove_duplicated_notes(midi.instruments[t].notes)  # remove possible duplicated notes
+            if len(midi.instruments[t].notes) == 0:
+                del midi.instruments[t]
+                continue
+            t += 1
+
+        # Recalculate max_tick is this could have change after notes quantization
+        if len(midi.instruments) > 0:
+            midi.max_tick = max([max([note.end for note in track.notes]) for track in midi.instruments])
+
+        if self.additional_tokens['Tempo']:
+            self.quantize_tempos(midi.tempo_changes, midi.ticks_per_beat)
+        # quantize_time_signatures(midi.time_signature_changes, midi.ticks_per_beat)
+
     def track_to_tokens(self, track: Instrument) -> List[Union[int, List[int]]]:
         """ Converts a track (miditoolkit.Instrument object) into a sequence of tokens
 
@@ -120,25 +142,25 @@ class MIDITokenizer:
         """
         raise NotImplementedError
 
-    def events_to_tokens(self, events: List[Event]) -> List[int]:
+    def _events_to_tokens(self, events: List[Event]) -> List[int]:
         """ Converts a list of Event objects into a list of tokens
-        You can override this method if necessary (e.g. CP Word)
+        You can override this method if necessary
 
         :param events: list of Events objects to convert
         :return: list of corresponding tokens
         """
-        return [self.event2token[f'{event.name}_{event.value}'] for event in events]
+        return [self.vocab.event_to_token[str(event)] for event in events]
 
-    def tokens_to_events(self, tokens: List[int]) -> List[Event]:
+    def _tokens_to_events(self, tokens: List[int]) -> List[Event]:
         """ Convert a sequence of tokens in their respective event objects
-        You can override this method if necessary (e.g. CP Word)
+        You can override this method if necessary
 
         :param tokens: sequence of tokens to convert
         :return: the sequence of corresponding events
         """
         events = []
         for token in tokens:
-            name, val = self.token2event[token].split('_')
+            name, val = self.vocab.token_to_event[token].split('_')
             events.append(Event(name, None, val, None))
         return events
 
@@ -186,31 +208,110 @@ class MIDITokenizer:
         """
         raise NotImplementedError
 
-    def _create_vocabulary(self, program_tokens: bool) -> Tuple[dict, dict, dict]:
-        """ Create the tokens <-> event dictionaries
-        These dictionaries are created arbitrary according to constants defined
-        at the top of this file.
-        Note that when using them (prepare_data method), there is no error-handling
-        so you must be sure that every case is covered by the dictionaries.
-        NOTE: token index 0 is often used as a padding index during training, it might
-        be preferable to leave it as it to pad your batch sequences
+    def quantize_notes(self, notes: List[Note], time_division: int, pitch_range: range = None):
+        """ Quantize the notes items, i.e. their pitch, velocity, start and end values.
+        It shifts the notes so they start at times that match the quantization (e.g. 16 samples per bar)
+        Notes with pitches outside of self.pitch_range will simply be deleted.
 
-        :param program_tokens: creates tokens for MIDI programs in the dictionary
-        :return: the dictionaries, one for each translation
+        :param notes: notes to quantize
+        :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
+        :param pitch_range: pitch range from within notes should be (default None -> self.pitch_range)
+        """
+        if pitch_range is None:
+            pitch_range = self.pitch_range
+        ticks_per_sample = int(time_division / max(self.beat_res.values()))
+        i = 0
+        while i < len(notes):
+            if notes[i].pitch not in pitch_range:
+                del notes[i]
+                continue
+            start_rest = notes[i].start % ticks_per_sample
+            end_rest = notes[i].end % ticks_per_sample
+            notes[i].start += -start_rest if start_rest <= ticks_per_sample / 2 else ticks_per_sample - start_rest
+            notes[i].end += -end_rest if end_rest <= ticks_per_sample / 2 else ticks_per_sample - end_rest
+
+            if notes[i].start == notes[i].end:  # if this happens to often, consider using a higher beat resolution
+                notes[i].end += ticks_per_sample  # like 8 samples per beat or 24 samples per bar
+
+            notes[i].velocity = min(self.velocities, key=lambda x: abs(x - notes[i].velocity))
+            i += 1
+
+    def quantize_tempos(self, tempos: List[TempoChange], time_division: int):
+        """ Quantize the times and tempo values of tempo change events.
+        Consecutive identical tempo changes will be removed.
+
+        :param tempos: tempo changes to quantize
+        :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
+        """
+        ticks_per_sample = int(time_division / max(self.beat_res.values()))
+        prev_tempo = -1
+        i = 0
+        while i < len(tempos):
+            # Quantize tempo value
+            tempos[i].tempo = min(self.tempos, key=lambda x: abs(x - tempos[i].tempo))
+            if tempos[i].tempo == prev_tempo:
+                del tempos[i]
+                continue
+            rest = tempos[i].time % ticks_per_sample
+            tempos[i].time += -rest if rest <= ticks_per_sample / 2 else ticks_per_sample - rest
+            prev_tempo = tempos[i].tempo
+            i += 1
+
+    @staticmethod
+    def quantize_time_signatures(time_sigs: List[TimeSignature], time_division: int):
+        """ Quantize the time signature changes, delayed to the next bar.
+        See MIDI 1.0 Detailed specifications, pages 54 - 56, for more information on
+        delayed time signature messages.
+
+        :param time_sigs: time signature changes to quantize
+        :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
+        """
+        ticks_per_bar = time_division * time_sigs[0].numerator
+        current_bar = 0
+        previous_tick = 0  # first time signature change is always at tick 0
+        for time_sig in time_sigs[1:]:
+            # determine the current bar of time sig
+            bar_offset, rest = divmod(time_sig.time - previous_tick, ticks_per_bar)
+            if rest > 0:  # time sig doesn't happen on a new bar, we update it to the next bar
+                bar_offset += 1
+                time_sig.time = previous_tick + bar_offset * ticks_per_bar
+
+            # Update values
+            ticks_per_bar = time_division * time_sig.numerator
+            current_bar += bar_offset
+            previous_tick = time_sig.time
+
+    def add_sos_eos_to_seq(self, seq: List[int]):
+        """ Adds Start Of Sequence (SOS) and End Of Sequence EOS tokens to a sequence of tokens:
+        SOS at the beginning, EOS at the end.
+
+        :param seq: sequence of tokens
+        """
+        seq.insert(0, self.vocab['SOS_None'])
+        seq.append(self.vocab['EOS_None'])
+
+    def _create_vocabulary(self, *args, **kwargs) -> Vocabulary:
+        """ Creates the Vocabulary object of the tokenizer.
+        See the docstring of the Vocabulary class for more details about how to use it.
+        NOTE: token index 0 is often used as a padding index during training
+        NOTE 2: SOS and EOS tokens should be set to -1 and -2 respectively.
+                use Vocabulary.add_sos_eos_to_vocab to add them
+
+        :return: the vocabulary object
         """
         raise NotImplementedError
 
-    def _create_token_types_graph(self) -> Dict[str, List[str]]:
+    def create_token_types_graph(self) -> Dict[str, List[str]]:
         """ Creates a dictionary for the directions of the token types of the encoding"""
         raise NotImplementedError
 
-    def _create_durations_tuples(self) -> List[Tuple]:
-        """ Creates the possible durations in bar / beat units, as tuple of the form:
-        (beat, pos, res) where beat is the number of beats, pos the number of "frames"
-        ans res the beat resolution considered (frames per beat)
+    def __create_durations_tuples(self) -> List[Tuple]:
+        """ Creates the possible durations in beat / position units, as tuple of the form:
+        (beat, pos, res) where beat is the number of beats, pos the number of "samples"
+        ans res the beat resolution considered (samples per beat)
         Example: (2, 5, 8) means the duration is 2 beat long + position 5 / 8 of the ongoing beat
         In pure ticks we have: duration = (beat * res + pos) * time_division // res
-            Is equivalent to: duration = nb_of_frames * ticks_per_frame
+            Is equivalent to: duration = nb_of_samples * ticks_per_sample
         So in the last example, if time_division is 384: duration = (2 * 8 + 5) * 384 // 8 = 1008 ticks
 
         :return: the duration bins
@@ -222,10 +323,36 @@ class MIDITokenizer:
         del durations[0]  # removes duration of 0
         return durations
 
+    def __create_rests(self) -> List[Tuple]:
+        """ Creates the possible rests in beat / position units, as tuple of the form:
+        (beat, pos) where beat is the number of beats, pos the number of "samples"
+        The rests are calculated from the value of self.additional_tokens[rest_range],
+        which first value divide a beat to determine the minimum rest represented,
+        and the second the maximum rest in beats.
+        The rests shorter than 1 beat will scale x2, as rests in music theory (semiquaver, quaver, crotchet...)
+        Note that the values of the rests in positions will be determined by the beat
+        resolution of the first range (self.beat_res)
+
+        Example: (4, 6) and a first beat resolution of 8 will give the rests:
+            [(0, 2), (0, 4), (1, 0), (2, 0), (3, 0), (4, 0), (5, 0), (6, 0)]
+
+        :return: the rests
+        """
+        div, max_beat = self.additional_tokens['rest_range']
+        assert div % 2 == 0 and div <= self._first_beat_res, \
+            f'The minimum rest must be divisible by 2 and lower than the first beat resolution ({self._first_beat_res})'
+        rests = []
+        while div > 1:
+            rests.append((0, self._first_beat_res // div))
+            div //= 2
+        rests += [(i, 0) for i in range(1, max_beat + 1)]
+        return rests
+
     def tokenize_midi_dataset(self, midi_paths: Union[List[str], List[Path], List[PurePath]],
                               out_dir: Union[str, Path, PurePath], validation_fn: Callable[[MidiFile], bool] = None,
                               logging: bool = True):
         """ Converts a dataset / list of MIDI files, into their token version and save them as json files
+        NOTE: MIDIs with a time division lower than 4 times the beat resolution will be discarded.
 
         :param midi_paths: paths of the MIDI files
         :param out_dir: output directory to save the converted files
@@ -234,11 +361,11 @@ class MIDITokenizer:
         :param logging: logs a progress bar
         """
         Path(out_dir).mkdir(parents=True, exist_ok=True)
-        
+
         # Making a directory of the parent folders for the JSON file
         # parent_dir = PurePath(midi_paths[0]).parent[0]
         # PurePath(out_dir, parent_dir).mkdir(parents=True, exist_ok=True)
-            
+
         for m, midi_path in enumerate(midi_paths):
             if logging:
                 bar_len = 60
@@ -255,6 +382,9 @@ class MIDITokenizer:
             except Exception as _:  # ValueError, OSError, FileNotFoundError, IOError, EOFError, mido.KeySignatureError
                 continue
 
+            # Checks the time division is valid
+            if midi.ticks_per_beat < max(self.beat_res.values()) * 4:
+                continue
             # Passing the MIDI to validation tests if given
             if validation_fn is not None:
                 if not validation_fn(midi):
@@ -264,11 +394,33 @@ class MIDITokenizer:
             tokens = self.midi_to_tokens(midi)
             midi_programs = get_midi_programs(midi)
             midi_name = PurePath(midi_path).stem
-            with open(PurePath(out_dir, midi_name).with_suffix(".json"), 'w') as outfile:
-                json.dump([tokens, midi_programs], outfile)
+            self.save_tokens(tokens, PurePath(out_dir, midi_name).with_suffix(".json"), midi_programs)
 
         self.save_params(out_dir)  # Saves the parameters with which the MIDIs are converted
 
+    @staticmethod
+    def save_tokens(tokens, path: Union[str, Path, PurePath], programs=None):
+        """ Saves tokens as a JSON file.
+
+        :param tokens: tokens, as any format
+        :param path: path of the file to save
+        :param programs: (optional), programs of the associated tokens, should be
+                        given as a tuples (int, bool) for (program, is_drum)
+        """
+        with open(path, 'w') as outfile:
+            json.dump([tokens, programs] if programs is not None else [tokens], outfile)
+
+    @staticmethod
+    def load_tokens(path: Union[str, Path, PurePath]) -> Tuple[Any, Any]:
+        """ Loads tokens saved as JSON files.
+
+        :param path: path of the file to load
+        :return: the tokens, with the associated programs if saved with
+        """
+        with open(path) as file:
+            data = json.load(file)
+            return data[0], data[1] if len(data) > 1 else None
+
     def save_params(self, out_dir: Union[str, Path, PurePath]):
         """ Saves the base parameters of this encoding in a txt file
         Useful to keep track of how a dataset has been tokenized / encoded
@@ -282,7 +434,7 @@ class MIDITokenizer:
         with open(PurePath(out_dir, 'config').with_suffix(".txt"), 'w') as outfile:
             json.dump({'pitch_range': (self.pitch_range.start, self.pitch_range.stop),
                        'beat_res': {f'{k1}_{k2}': v for (k1, k2), v in self.beat_res.items()},
-                       'nb_velocities': len(self.velocity_bins),
+                       'nb_velocities': len(self.velocities),
                        'additional_tokens': self.additional_tokens,
                        'encoding': self.__class__.__name__}, outfile, indent=4)
 
@@ -314,83 +466,71 @@ def get_midi_programs(midi: MidiFile) -> List[Tuple[int, bool]]:
     return [(int(track.program), track.is_drum) for track in midi.instruments]
 
 
-def quantize_note_times(notes: List[Note], time_division: int, beat_res: int):
-    """ Quantize the notes items start and end values.
-    It shifts the notes so they start at times that match the quantization (e.g. 16 frames per bar)
-
-    :param notes: notes to quantize
-    :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
-    :param beat_res: number of frames (time steps, or positions) per beat
-    """
-    ticks = int(time_division / beat_res)
-    quantized_ticks = np.arange(0, max([n.end for n in notes]) + 2 * ticks, ticks, dtype=int)
-    for i, note in enumerate(notes):  # items are notes
-        note.start = quantized_ticks[np.argmin(np.abs(quantized_ticks - note.start))]
-        note.end = quantized_ticks[np.argmin(np.abs(quantized_ticks - note.end))]
-
-        if note.start == note.end:  # if this happens to often, consider using a higher beat resolution
-            note.end += ticks  # like 8 frames per beat or 24 frames per bar
-
-
-def quantize_tempos(tempos: List[TempoChange], time_division: int, beat_res: int):
-    """ Quantize the times of tempo change events.
-
-    :param tempos: tempo changes to quantize
-    :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
-    :param beat_res: number of frames (time steps, or positions) per beat
-    """
-    ticks = int(time_division / beat_res)
-    quantized_ticks = np.arange(0, max([t.time for t in tempos]) + 2 * ticks, ticks, dtype=int)
-    for tempo_change in tempos:
-        tempo_change.time = quantized_ticks[np.argmin(np.abs(quantized_ticks - tempo_change.time))]
-
-
 def remove_duplicated_notes(notes: List[Note]):
     """ Remove possible duplicated notes, i.e. with the same pitch, starting and ending times.
-    Before running this function make sure the notes has been sorted by start and pitch:
-    notes.sort(key=lambda x: (x.start, x.pitch))
+    Before running this function make sure the notes has been sorted by start then pitch then end values:
+    notes.sort(key=lambda x: (x.start, x.pitch, x.end))
 
     :param notes: notes to analyse
     """
     for i in range(len(notes) - 1, 0, -1):  # removing possible duplicated notes
         if notes[i].pitch == notes[i - 1].pitch and notes[i].start == notes[i - 1].start and \
-                notes[i].end == notes[i - 1].end:
+                notes[i].end >= notes[i - 1].end:
             del notes[i]
 
 
-def detect_chords(notes: List[Note], time_division: int) -> List[Event]:
+def detect_chords(notes: List[Note], time_division: int, beat_res: int = 4, onset_offset: int = 1,
+                  only_known_chord: bool = False, simul_notes_limit: int = 20) -> List[Event]:
     """ Chord detection method.
-    NOTE: on very large tracks with high note density this method can be very slow !
+    NOTE: make sure to sort notes by start time then pitch before: notes.sort(key=lambda x: (x.start, x.pitch))
+    NOTE2: on very large tracks with high note density this method can be very slow !
     If you plan to use it with the Maestro or GiantMIDI datasets, it can take up to
     hundreds of seconds per MIDI depending on your cpu.
     One time step at a time, it will analyse the notes played together
     and detect possible chords.
 
-    :param notes: notes to analyse
+    :param notes: notes to analyse (sorted by starting time, them pitch)
     :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
+    :param beat_res: beat resolution, i.e. nb of samples per beat (default 4)
+    :param onset_offset: maximum offset (in samples) ∈ N separating notes starts to consider them
+                            starting at the same time / onset (default is 1)
+    :param only_known_chord: will select only known chords. If set to False, non recognized chords of
+                            n notes will give a chord_n event (default False)
+    :param simul_notes_limit: nb of simultaneous notes being processed when looking for a chord
+            this parameter allows to speed up the chord detection (default 20)
     :return: the detected chords as Event objects
     """
+    assert simul_notes_limit >= 5, 'simul_notes_limit must be higher than 5, chords can be made up to 5 notes'
     tuples = []
     for note in notes:
         tuples.append((note.pitch, int(note.start), int(note.end)))
     notes = np.asarray(tuples)
 
+    time_div_half = time_division // 2
+    onset_offset = time_division * onset_offset / beat_res
+
     count = 0
+    previous_tick = -1
     chords = []
     while count < len(notes):
-        # Gather the notes around the same time step
-        onset_notes = notes[count:]
-        onset_notes = onset_notes[np.where(onset_notes[:, 1] <= notes[count][1] + time_division / 4)]
+        # Checks we moved in time after last step, otherwise discard this tick
+        if notes[count, 1] == previous_tick:
+            count += 1
+            continue
+
+        # Gathers the notes around the same time step
+        onset_notes = notes[count:count + simul_notes_limit]  # reduces the scope
+        onset_notes = onset_notes[np.where(onset_notes[:, 1] <= onset_notes[0, 1] + onset_offset)]
 
         # If it is ambiguous, e.g. the notes lengths are too different
-        if np.any(np.abs(onset_notes[:, 2] - onset_notes[0, 2]) > time_division / 2):
+        if np.any(np.abs(onset_notes[:, 2] - onset_notes[0, 2]) > time_div_half):
             count += len(onset_notes)
             continue
 
         # Selects the possible chords notes
-        if notes[count][2] - notes[count][1] <= time_division / 2:
-            onset_notes = onset_notes[np.where(onset_notes[:, 1] == onset_notes[0][1])]
-        chord = onset_notes[np.where(onset_notes[:, 2] - onset_notes[0, 2] <= time_division / 2)]
+        if notes[count, 2] - notes[count, 1] <= time_div_half:
+            onset_notes = onset_notes[np.where(onset_notes[:, 1] == onset_notes[0, 1])]
+        chord = onset_notes[np.where(onset_notes[:, 2] - onset_notes[0, 2] <= time_div_half)]
 
         # Creates the "chord map" and see if it has a "known" quality, append a chord event if it is valid
         chord_map = (chord[:, 0] - chord[0, 0]).tolist()
@@ -400,7 +540,11 @@ def detect_chords(notes: List[Note], time_division: int) -> List[Event]:
                 if known_chord == chord_map:
                     chord_quality = quality
                     break
+            if only_known_chord and isinstance(chord_quality, int):
+                count += len(onset_notes)  # Move to the next notes
+                continue  # this chords was not recognize and we don't want it
             chords.append((chord_quality, min(chord[:, 1]), chord_map))
+        previous_tick = max(onset_notes[:, 1])
         count += len(onset_notes)  # Move to the next notes
 
     events = []
@@ -409,47 +553,11 @@ def detect_chords(notes: List[Note], time_division: int) -> List[Event]:
     return events
 
 
-def _detect_chords_python(notes: List[Note], time_division: int) -> List[Event]:
-    """ DEPRECIATED
-    Old chord detection method, equivalent to detect_chords but 100% python
-    The code is more elegant but slower
-
-    :param notes: notes to analyse
-    :param time_division: MIDI time division / resolution, in ticks/beat (of the MIDI being parsed)
-    :return: the detected chords as Event objects
-    """
-    count = 0
-    chords = []
-    while count < len(notes):
-        # Gather the notes around the same time step
-        # onset_notes = [note for note in notes if abs(note.start - notes[count].start) <= time_division / 4]
-        onset_notes = [note for note in notes[count:] if note.start - notes[count].start <= time_division / 4]
-
-        # If it is ambiguous, e.g. the notes lengths are too different
-        if any(abs(note.end - notes[count].end) > time_division / 2 for note in onset_notes):
-            count += len(onset_notes)
-            continue
-
-        # Selects the possible chords notes
-        if notes[count].end - notes[count].start <= time_division / 2:
-            onset_notes = [note for note in notes[count:] if note.start == notes[count].start]
-        chord = [note for note in onset_notes if abs(note.end - notes[count].end) <= time_division / 2]
-
-        # Creates the "chord map" and see if it has a "known" quality, append a chord event if it is valid
-        chord_map = [chord[i].pitch - chord[0].pitch for i in range(len(chord))]
-        if 3 <= len(chord_map) <= 5 and chord_map[-1] <= 24:  # max interval between the root and highest degree
-            if chord_map in CHORD_MAPS.values():
-                chord_quality = list(CHORD_MAPS.keys())[list(CHORD_MAPS.values()).index(chord_map)]
-                chords.append(Event('Chord', min(note.start for note in chord), chord_quality, chord_map))
-            else:
-                chords.append(Event('Chord', min(note.start for note in chord), len(chord), chord_map))
-        count += len(onset_notes)  # Move to the next notes
-    return chords
-
-
 def merge_tracks(tracks: List[Instrument]) -> Instrument:
     """ Merge several miditoolkit Instrument objects
-    It will take the first object, and concat the notes of the others
+    All the tracks will be merged into the first Instrument object (notes concatenated and sorted),
+    beware of giving tracks with the same program (no assessment is performed)
+    The other tracks will be deleted.
 
     :param tracks: list of tracks to merge
     :return: the merged track
@@ -457,9 +565,33 @@ def merge_tracks(tracks: List[Instrument]) -> Instrument:
     tracks[0].name += ''.join([' / ' + t.name for t in tracks[1:]])
     tracks[0].notes = sum((t.notes for t in tracks), [])
     tracks[0].notes.sort(key=lambda note: note.start)
+    tracks = [tracks[0]]
     return tracks[0]
 
 
+def merge_same_program_tracks(tracks: List[Instrument]):
+    """ Takes a list of tracks and merge the ones with the same programs.
+    NOTE: Control change messages are not considered
+
+    :param tracks: list of tracks
+    """
+    # Gathers tracks programs and indexes
+    tracks_programs = [int(track.program) if not track.is_drum else -1 for track in tracks]
+
+    # Detects duplicated programs
+    duplicated_programs = [k for k, v in Counter(tracks_programs).items() if v > 1]
+
+    # Merges duplicated tracks
+    for program in duplicated_programs:
+        idx = [i for i in range(len(tracks)) if
+               (tracks[i].is_drum if program == -1 else tracks[i].program == program and not tracks[i].is_drum)]
+        tracks[idx[0]].name += ''.join([' / ' + tracks[i].name for i in idx[1:]])
+        tracks[idx[0]].notes = sum((tracks[i].notes for i in idx), [])
+        tracks[idx[0]].notes.sort(key=lambda note: (note.start, note.pitch))
+        for i in list(reversed(idx[1:])):
+            del tracks[i]
+
+
 def current_bar_pos(seq: List[int], bar_token: int, position_tokens: List[int], pitch_tokens: List[int],
                     chord_tokens: List[int] = None) -> Tuple[int, int, List[int], bool]:
     """ Detects the current state of a sequence of tokens
